@@ -13,6 +13,8 @@ import json
 
 import jsonschema
 
+import dtschema.dtb
+
 schema_base_url = "http://devicetree.org/"
 schema_basedir = os.path.dirname(os.path.abspath(__file__))
 
@@ -177,6 +179,26 @@ def _fixup_string_to_array(propname, subschema):
     subschema['items'] = [_extract_single_schemas(subschema)]
 
 
+def _fixup_reg_schema(propname, subschema):
+    # nothing to do if we don't have a set of string schema
+    if propname != 'reg':
+        return
+
+    if 'items' in subschema:
+        if isinstance(subschema['items'], list):
+            item_schema = subschema['items'][0]
+        else:
+            item_schema = subschema['items']
+        if not _is_int_schema(item_schema):
+            return
+    elif _is_int_schema(subschema):
+        item_schema = subschema
+    else:
+        return
+
+    subschema['items'] = [ {'items': [ _extract_single_schemas(item_schema) ] } ]
+
+
 def _is_matrix_schema(subschema):
     if 'items' not in subschema:
         return False
@@ -190,10 +212,27 @@ def _is_matrix_schema(subschema):
 
     return False
 
+# If we have a matrix with variable inner and outer dimensions, then drop the dimensions
+# because we have no way to reconstruct them.
+def _fixup_int_matrix(subschema):
+    if not _is_matrix_schema(subschema):
+        return
+
+    outer_dim = _get_array_range(subschema)
+    inner_dim = _get_array_range(subschema.get('items', {}))
+
+    if outer_dim[0] != outer_dim[1] and inner_dim[0] != inner_dim[1]:
+        subschema.pop('items', None)
+        subschema.pop('maxItems', None)
+        subschema.pop('minItems', None)
+        subschema['type'] = 'array'
+
 
 int_array_re = re.compile('int(8|16|32|64)-array')
 unit_types_re = re.compile('-(bits|percent|mhz|hz|sec|ms|us|ns|ps|mm|microamp|microamp-hours|ohms|micro-ohms|microwatt-hours|microvolt|picofarads|celsius|millicelsius|kpascal)$')
 
+# Remove this once we remove array to matrix fixups
+known_array_props = {'assigned-clock-rates', 'linux,keycodes'}
 
 def is_int_array_schema(propname, subschema):
     if 'allOf' in subschema:
@@ -206,7 +245,7 @@ def is_int_array_schema(propname, subschema):
                 return int_array_re.search(item['$ref'])
     if '$ref' in subschema:
         return int_array_re.search(subschema['$ref'])
-    elif unit_types_re.search(propname):
+    elif unit_types_re.search(propname) or propname in known_array_props:
         return True
 
     return 'items' in subschema and \
@@ -375,7 +414,9 @@ def fixup_vals(propname, schema):
 
     schema.pop('description', None)
 
+    _fixup_reg_schema(propname, schema)
     _fixup_remove_empty_items(schema)
+    _fixup_int_matrix(schema)
     _fixup_int_array_min_max_to_matrix(propname, schema)
     _fixup_int_array_items_to_matrix(propname, schema)
     _fixup_string_to_array(propname, schema)
@@ -707,7 +748,155 @@ def process_schemas(schema_paths, core_schema=True):
     return schemas
 
 
+def _get_array_range(subschema):
+    if isinstance(subschema, list):
+        if len(subschema) != 1:
+            return (0, 0)
+        subschema = subschema[0]
+    if 'items' in subschema and isinstance(subschema['items'], list):
+        max = len(subschema['items'])
+        min = subschema.get('minItems', max)
+    else:
+        min = subschema.get('minItems', 1)
+        max = subschema.get('maxItems', subschema.get('minItems', 0))
+
+    return (min, max)
+
+
+def _merge_dim(dim1, dim2):
+    d = []
+    for i in range(0, 2):
+        if dim1[i] == (0, 0):
+            d.insert(i, dim2[i])
+        elif dim2[i] == (0, 0):
+            d.insert(i, dim1[i])
+        else:
+            d.insert(i, (min(dim1[i] + dim2[i]), max(dim1[i] + dim2[i])))
+
+    return tuple(d)
+
+
+type_re = re.compile('(flag|u?int(8|16|32|64)(-(array|matrix))?|string(-array)?|phandle(-array)?)')
+
+
+def _extract_prop_type(props, schema, propname, subschema):
+    if not isinstance(subschema, dict):
+        return
+
+    if subschema.keys() & {'properties', 'patternProperties', 'additionalProperties'}:
+        _extract_subschema_types(props, schema, subschema)
+
+    # We only support local refs
+    if '$ref' in subschema and subschema['$ref'].startswith('#/'):
+        sch_path = subschema['$ref'].split('/')[1:]
+        subschema = schema
+        for p in sch_path:
+            subschema = subschema[p]
+        #print(propname, sch_path, subschema, file=sys.stderr)
+        _extract_prop_type(props, schema, propname, subschema)
+
+    try:
+        prop_type = type_re.search(subschema['$ref']).group(0)
+    except:
+        if 'type' in subschema and subschema['type'] == 'boolean':
+            prop_type = 'flag'
+        elif 'items' in subschema:
+            items = subschema['items']
+            if (isinstance(items, list) and _is_string_schema(items[0])) or \
+               (isinstance(items, dict) and _is_string_schema(items)):
+                # implicit string type
+                prop_type = 'string-array'
+            else:
+                prop_type = None
+        else:
+            prop_type = None
+
+    if prop_type:
+        props.setdefault(propname, {})
+
+        if prop_type == 'phandle-array' or prop_type.endswith('-matrix'):
+            dim = (_get_array_range(subschema), _get_array_range(subschema.get('items', {})))
+            if dim != ((0, 0), (0, 0)):
+                if not 'dim' in props[propname]:
+                    props[propname]['dim'] = dim
+                elif props[propname]['dim'] != dim:
+                    # Conflicting dimensions
+                    props[propname]['dim'] = _merge_dim(props[propname]['dim'], dim)
+
+        if 'type' in props[propname]:
+            if prop_type in props[propname]['type']:
+                # Already have the same type
+                return
+            for t in props[propname]['type']:
+                if prop_type in t:
+                    # Already have the looser type
+                    return
+                if t in prop_type:
+                    # Replace scalar type with array type
+                    i = props[propname]['type'].index(t)
+                    props[propname]['type'][i] = prop_type
+                    props[propname]['$id'][i] = schema['$id']
+                    return
+        else:
+            props[propname]['type'] = []
+            props[propname]['$id'] = []
+
+        props[propname]['type'] += [ prop_type ]
+        props[propname]['$id'] += [ schema['$id'] ]
+        return
+
+    for k in subschema.keys() & {'allOf', 'oneOf', 'anyOf'}:
+        for v in subschema[k]:
+            _extract_prop_type(props, schema, propname, v)
+
+
+def _extract_subschema_types(props, schema, subschema):
+    if not isinstance(subschema, dict):
+        return
+
+    for k in subschema.keys() & {'properties', 'patternProperties', 'additionalProperties'}:
+        if isinstance(subschema[k], dict):
+            for p,v in subschema[k].items():
+                _extract_prop_type(props, schema, p, v)
+
+
+def extract_types(schemas):
+    if not isinstance(schemas, list):
+        return
+
+    props = {}
+    for sch in schemas:
+        _extract_subschema_types(props, sch, sch)
+
+    return props
+
+
+def get_prop_types():
+    global schema_cache
+    pat_props = {}
+
+    props = dtschema.extract_types(schema_cache)
+
+    # hack to remove aliases pattern
+    del props['^[a-z][a-z0-9\-]*$']
+
+    # Remove all properties without a type
+    for key in [key for key in props if 'type' not in props[key] ]: del props[key]
+
+    # Split out pattern properties
+    for key in [key for key in props if re.fullmatch('(^\^.*|.*\$$|.*[*[].*)', key) ]:
+        #print(key, props[key])
+        pat_props[key] = props[key]
+        pat_props[key]['regex'] = re.compile(key)
+        del props[key]
+
+    return [ props, pat_props ]
+
+
 def load(filename, line_number=False):
+    if filename.endswith('.dtb'):
+        with open(filename, 'rb') as f:
+            return [ dtschema.dtb.fdt_unflatten(f.read()) ]
     with open(filename, 'r', encoding='utf-8') as f:
         if line_number:
             return rtyaml.load(f.read())
@@ -758,12 +947,7 @@ def typeSize(validator, typeSize, instance, schema):
         yield jsonschema.ValidationError("size is %r, expected %r" % (size, typeSize))
 
 
-def phandle(validator, phandle, instance, schema):
-    if not isinstance(instance, sized_int) or not instance.phandle:
-        yield jsonschema.ValidationError("missing phandle tag in %r" % instance)
-
-
-DTVal = jsonschema.validators.extend(jsonschema.Draft201909Validator, {'typeSize': typeSize, 'phandle': phandle})
+DTVal = jsonschema.validators.extend(jsonschema.Draft201909Validator, {'typeSize': typeSize})
 
 
 class DTValidator(DTVal):
